#
#  Type: Loaded File
#
#  This type stores a loaded file, to support the operations
#  - parse it into a parsed file
#  - modify it and write it out again, mostly unchanged
#
#  A loaded file consists of a list of items. Each item is itself a list.
#  [ raw, "text" ]              raw text, e.g. a comment
#  [ begin, "name" ]            begin a group
#  [ end ]                      end a group
#  [ set, type, "name", "val"]  set a variable. type=1 to make it expandable
#  [ app, type, "name", "val"]  append to a variable
#  [ rule, "lhs", "rhs" ]       a rule header
#  [ pat, "lhs", "rhs" ]        a pattern rule header
#  [ code, "text" ]             code for a rule
#  [ include, "foo" ]           include
#
#  For 'raw', the text includes the newline (which means [raw,""] is a no-op),
#  for all the others, the formatter has to add one.
#

package LoadedFile;

use Util;
use strict;

# Given a file name, load that file, and return loaded file.
sub load($$) {
    my $proto = shift;
    my $filename = shift;
    open FILE, "< $filename" or
      diag_perror($filename);
    diag_info("load: loading '$filename'...");

    my $accept_code = 0;
    my $group_level = 0;
    my $pdata = [];
    my $result = { name => $filename,
                   data => $pdata };
    bless($result, ref($proto) || $proto);
    my $line_nr = 0;

    while (defined(my $line = <FILE>)) {
        # Read a physical line
        my $this_line_nr = ++$line_nr;
        $line = util_chomp($line);
        while ($line =~ s/\\$//) {
            my $next = <FILE>;
            if (!defined($next)) { diag_error("$filename:$this_line_nr: ends with backslashed line") }
            $line .= util_chomp($next);
            ++$line_nr;
        }

        # What is it?
        if ($line =~ /^\s*$/ || $line =~ /^\s*\#/) {
            push @$pdata, [ 'raw', "$line\n" ];
        } elsif ($line =~ /^begin (.*)/) {
            push @$pdata, [ 'begin', $1 ];
            ++$group_level;
            $accept_code = 0;
        } elsif ($line =~ /^end$/) {
            push @$pdata, [ 'end' ];
            if (!$group_level) { diag_error("$filename:$this_line_nr: 'end' not allowed here") }
            --$group_level;
            $accept_code = 0;
        } elsif ($line =~ /^(include)\s+(.*)/) {
            push @$pdata, [ $1, $2 ];
        } elsif ($line =~ /^(\S+)\s*(|\+:?|:\+?)=\s*(.*)/) {
            my $name = $1;
            my $type = $2;
            my $value = $3;
            my $isAppend = $type =~ /\+/;
            my $isExpand = $type =~ /:/;
            if ($isAppend) {
                push @$pdata, [ 'app', $isExpand, $name, $value ];
            } else {
                push @$pdata, [ 'set', $isExpand, $name, $value ];
            }
            $accept_code = 0;
        } elsif ($line =~ /^\.(.*?)\.(.*?):$/) {
            push @$pdata, [ 'pat', $1, $2 ];
            $accept_code = 1;
        } elsif ($line =~ /^(\S.*?):(.*)/) {
            push @$pdata, [ 'rule', util_trim($1), util_trim($2) ];
            $accept_code = 1;
        } elsif ($line =~ /^\s+(.*)/) {
            if (!$accept_code) { diag_error("$filename:$this_line_nr: code not allowed here") }
            push @$pdata, [ 'code', $1 ];
        } else {
            diag_error("$filename:$this_line_nr: parse error")
        }
    }
    if ($group_level) { diag_error("$filename: $group_level 'begin' not closed at end") }
    close FILE;
    $result;
}

sub clone( $ ) {
    my $self = shift;
    my $newself = { name => $self->{name},
                    data => [ @{$self->{data}} ] };
    bless $newself;
}

sub getData() {
    my $self = shift;
    $self->{data};
}

sub getName() {
    my $self = shift;
    $self->{name};
}

sub _stringValue ( $$ ) {
    my $line = shift;
    my $fail = shift;

    my $id = $line->[0];
    if ($id eq 'raw') {
        $line->[1];
    } elsif ($id eq 'begin') {
        "begin " . $line->[1] . "\n";
    } elsif ($id eq 'end') {
        "end\n";
    } elsif ($id eq 'include') {
        "include " . $line->[1] . "\n";
    } elsif ($id eq 'app') {
        if ($line->[1]) {
            util_format_variable($line->[2], "+:=", $line->[3]);
        } else {
            util_format_variable($line->[2], "+=", $line->[3]);
        }
    } elsif ($id eq 'set') {
        if ($line->[1]) {
            util_format_variable($line->[2], ":=", $line->[3]);
        } else {
            util_format_variable($line->[2], "=", $line->[3]);
        }
    } elsif ($id eq 'pat') {
        "." . $line->[1], "." . $line->[2], ":\n";
    } elsif ($id eq 'rule') {
        $line->[1] . ": " . $line->[2] . "\n";
    } elsif ($id eq 'code') {
        "\t" . $line->[1] . "\n";
    } else {
        if ($fail) {
            die "invalid '$id' in LoadedFile";
        } else {
            "<error>"
        }
    }
}

sub setLine( $$$ ) {
    my ($self, $line, $value) = @_;
    $self->{data}[$line-1] = $value;
    if (cmd_bool('dry', 'n')) {
        my $str = _stringValue($value, 0);
        $str =~ s/^/  /;
        print "Changing block $line of $self->{name} to:\n$str\n";
    }
}

sub addLine( $$ ) {
    my ($self, $value) = @_;
    push @{$self->{data}}, $value;
    if (cmd_bool('dry', 'n')) {
        my $str = _stringValue($value, 0);
        $str =~ s/^/  /;
        print "Adding to $self->{name}:\n$str\n";
    }
}

sub write ( $$ ) {
    my ($self, $fileName) = @_;
    open FILE, "> $fileName"
      or diag_perror($fileName);

    foreach my $line (@{$self->{data}}) {
        print FILE _stringValue($line, 1);
#         my $id = $line->[0];
#         if ($id eq 'raw') {
#             print FILE $line->[1];
#         } elsif ($id eq 'begin') {
#             print FILE "begin ", $line->[1], "\n";
#         } elsif ($id eq 'end') {
#             print FILE "end\n";
#         } elsif ($id eq 'include') {
#             print FILE "include ", $line->[1], "\n";
#         } elsif ($id eq 'app') {
#             print FILE util_format_variable($line->[2], "+=", $line->[3]);
#         } elsif ($id eq 'set') {
#             if ($line->[1]) {
#                 print FILE util_format_variable($line->[2], ":=", $line->[3]);
#             } else {
#                 print FILE util_format_variable($line->[2], "=", $line->[3]);
#             }
#         } elsif ($id eq 'pat') {
#             print FILE ".", $line->[1], ".", $line->[2], ":\n";
#         } elsif ($id eq 'rule') {
#             print FILE $line->[1], ": ", $line->[2], "\n";
#         } elsif ($id eq 'code') {
#             print FILE "\t", $line->[1], "\n";
#         } else {
#             die "invalid '$id' in LoadedFile";
#         }
    }
    close FILE;
}

1;
